/*
 * Copyright (c) 2012 AgroSense.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    AgroSense - initial API and implementation and/or initial documentation
 */
package nl.cloudfarming.client.isobus.export;

import com.vividsolutions.jts.geom.Geometry;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.logging.Logger;
import javax.swing.ImageIcon;
import javax.swing.JFileChooser;
import nl.cloudfarming.client.field.model.ActivityField;
import nl.cloudfarming.client.field.model.TreatmentZone;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DataLog;
import nl.cloudfarming.client.isobus.model.GridTypeEnum;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import nl.cloudfarming.client.isobus.model.IsobusMarshaller;
import nl.cloudfarming.client.isobus.model.IsobusPreference;
import nl.cloudfarming.client.isobus.model.IsobusTask;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import nl.cloudfarming.client.isobus.model.IsobusTaskStatusEnum;
import nl.cloudfarming.client.isobus.model.IsobusTreatmentZone;
import nl.cloudfarming.client.isobus.model.ProcessDataVariable;
import nl.cloudfarming.client.isobus.model.ValuePresentation;
import nl.cloudfarming.client.isobus.model.util.IsobusObjectFactoryUtil;
import nl.cloudfarming.client.isobus.type.IsobusManagerImpl;
import nl.cloudfarming.client.lib.geotools.GeometryTools;
import nl.cloudfarming.client.task.ExportAction;
import nl.cloudfarming.client.task.ExportableTask;
import nl.cloudfarming.client.task.TaskExporter;
import nl.cloudfarming.client.task.util.GeometryUtil;
import nl.cloudfarming.client.util.VersionUtil;
import org.apache.commons.lang.StringUtils;
import org.geotools.coverage.grid.GridCoverage2D;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.openide.awt.NotificationDisplayer;
import org.openide.filesystems.FileChooserBuilder;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.Exceptions;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;

/**
 *
 * @author Timon Veenstra <monezz@gmail.com>
 */
@NbBundle.Messages({"isobus exporter treatment zone limit exceeded=Maximum number of treatment zones exceeded, exporting first 254 treatment zones!",
    "isobus exporter generating isobus export=Exporting isobus",
    "isobus exporter generating isobus full description=Exporting of the isobus file is completed",
    "isobus exporter generating export done title=Export completed",
    "isobus exporter choose location=Choose export location",})
@ServiceProvider(service = TaskExporter.class, position = 1)
public class IsobusExporter implements TaskExporter {

    private static final String VERSION_MAJOR = "2";
    private static final String VERSION_MINOR = "0";
    // module name for dynamic lookup of MANAGEMENT_SOFTWARE_VERSION:
    private static final String MODULE_CODE_NAME = IsobusManagerImpl.class.getPackage().getName();
    private static final String TASK_CONTROLLER_VERSION = "1.0";
    private static final String MANAGEMENT_SOFTWARE_MANUFACTURER = "AgroSense";
    private static final Logger LOG = Logger.getLogger(IsobusExporter.class.getName());
    private static final short EMPTY_TREATMENTZONE_CODE = 0;
    private static final String EMPTY_TREATMENTZONE_DESIGNATOR = "Empty";
    private static final short EMPTY_TREATMENTZONE_COLOR = 0;
    private static final short DEFAULT_TREATMENTZONE_CODE = 1;
    private static final String DEFAULT_TREATMENTZONE_DESIGNATOR = "Default";
    private static final short DEFAULT_TREATMENTZONE_COLOR = 0;
    private static final String DATA_TRANSFER_ORIGIN = "1";
    private static final String VALUE_PRESENTATION_ID = "VPN1";
    private static final int VALUE_PRESENTATION_OFFSET = 0;
    private static final float VALUE_PRESENTATION_SCALE = 1.0f;
    private static final short VALUE_PRESENTATION_DECIMALS = 0;
    private static final String VALUE_PRESENTATION_DESIGNATOR = "aan/uit";
    private static final String ISOBUS_TASK_ID_PREFIX = "TSK";
    private static final IsobusTaskStatusEnum TASK_STATUS = IsobusTaskStatusEnum.INITIAL;
    private static final String GRID_FILE_EXT = "BIN";
    private static final String GRID_FILE_PREFIX = "GRD";
    private static final int GRID_ID_LENGTH = 5;
    private static final short TREATMENTZONE_TYPE_2_CODE = 2; // TZN-code to use for type 2 prototype tzn.
    private static final String TREATMENT_ZONE_TYPE_2_NAME = "Precision application";
    public static final String PROP_DDI = "DDI";
    public static final String PROP_WORK_WIDTH = "workWidth";
    public static final String PROP_MULTIPLIER = "multiplier";
    public static final String PROP_DEFAULT_TZN_VALUE= "defaultTznValue";
    private ImageIcon INFO_ICON = ImageUtilities.loadImageIcon("nl/cloudfarming/client/isobus/type/information.png", true);

    private ProcessDataVariable createProcessDataVariable(ValuePresentation presentation, DDI ddi) throws NumberFormatException {
        return IsobusObjectFactoryUtil.createProcessDataVariable(0.0, presentation, ddi, 1.0);
    }

    private IsobusTreatmentZone createEmptyTreatmentZone() {
        IsobusTreatmentZone emptyTreatmentZone = new IsobusTreatmentZone();
        emptyTreatmentZone.setTreatmentZoneCode(EMPTY_TREATMENTZONE_CODE);
        emptyTreatmentZone.setTreatmentZoneDesignator(EMPTY_TREATMENTZONE_DESIGNATOR);
        emptyTreatmentZone.setTreatmentZoneColor(EMPTY_TREATMENTZONE_COLOR);
        return emptyTreatmentZone;
    }

    private IsobusTreatmentZone createDefaultTreatmentZone() {
        IsobusTreatmentZone defaultTreatmentZone = new IsobusTreatmentZone();
        defaultTreatmentZone.setTreatmentZoneCode(DEFAULT_TREATMENTZONE_CODE);
        defaultTreatmentZone.setTreatmentZoneDesignator(DEFAULT_TREATMENTZONE_DESIGNATOR);
        defaultTreatmentZone.setTreatmentZoneColor(DEFAULT_TREATMENTZONE_COLOR);
        return defaultTreatmentZone;
    }
    
    private DataLog createDataLog() throws NumberFormatException {
        DataLog dataLog = new DataLog();
        dataLog.setDDI(new byte[] { Byte.parseByte("-33"), Byte.parseByte("-1") }); // "DFFF"
        dataLog.setMethod((short) 31);
        dataLog.setTimeInterval(1000);
        return dataLog;
    }

    private IsobusTask createIsobusTask(String taskId) {
        IsobusTask isobusTask = new IsobusTask();
        isobusTask.setTaskId(taskId);
        isobusTask.setTaskStatus(TASK_STATUS);
        return isobusTask;
    }

    //TODO: create a VPN based on a.o. the input source UOM,
    // the current hardcoded values only apply to the on/off DDI.
    private ValuePresentation createValuePresentation() {
        ValuePresentation presentation = new ValuePresentation();
        presentation.setValuePresentationId(VALUE_PRESENTATION_ID);
        presentation.setOffset(VALUE_PRESENTATION_OFFSET);
        presentation.setScale(VALUE_PRESENTATION_SCALE);
        presentation.setNumberOfDecimals(VALUE_PRESENTATION_DECIMALS);
        presentation.setUnitDesignator(VALUE_PRESENTATION_DESIGNATOR);
        return presentation;
    }

    private IsobusTaskData createIsobusTaskData() {
        // create root object IsobusTaskData and fill in static config data
        IsobusTaskData isobusTaskData = new IsobusTaskData();
        isobusTaskData.setVersionMajor(VERSION_MAJOR);
        isobusTaskData.setVersionMinor(VERSION_MINOR);
        isobusTaskData.setDataTransferOrigin(DATA_TRANSFER_ORIGIN);
        isobusTaskData.setManagementSoftwareManufacturer(MANAGEMENT_SOFTWARE_MANUFACTURER);
        // lookup installed version:
        isobusTaskData.setManagementSoftwareVersion(VersionUtil.getVersion(MODULE_CODE_NAME));
        isobusTaskData.setTaskControllerVersion(TASK_CONTROLLER_VERSION);
        return isobusTaskData;
    }

    /**
     * <p>Create one or more {@link IsobusTreatmentZone} from the provided list
     * of {@link TreatmentZone}</p> <p>Can create list for both type of
     * {@link GridTypeEnum}. For {@link GridTypeEnum#TYPE_1} creates a
     * {@link IsobusTreatmentZone} with one {@link ProcessDataVariable} for
     * every {@link TreatmentZone} in the passed list. For
     * {@link GridTypeEnum#TYPE_2} will create one {@link IsobusTreatmentZone}
     * with a {@link ProcessDataVariable} for every {@link TreatmentZone} in the
     * passed list</p>
     *
     * @param treatmentZones
     * @param presentation
     * @param grid
     * @param gridType
     * @return
     */
    protected List<IsobusTreatmentZone> createTreatmentZones(List<TreatmentZone> treatmentZones, ValuePresentation presentation, TreatmentZoneGrid grid, GridTypeEnum gridType, DDI ddi, double multiplier) {
        List<IsobusTreatmentZone> created = new ArrayList<>();

        switch (gridType) {
            case TYPE_1:
                // loop through treatmentzones in an agrosense task, convert them to 
                // isobus treatmentzones and add them to the task object (via TZN node)
                int counter = 2;
                for (TreatmentZone treatmentZone : treatmentZones) {
                    IsobusTreatmentZone isobusTreatmentZone = IsobusObjectFactoryUtil.createIsobusTreatmentZone(presentation, treatmentZone, grid.getTreatmentZoneCode(treatmentZone), ddi, multiplier);
                    if (isobusTreatmentZone != null) {
                        created.add(isobusTreatmentZone);
                    }
                    //FIXME: ISOBUS supports a maximum number of 255  treatment zones 
                    // (including the initial empty one, and the default one). In order to support more 
                    // treatment zones they should probably be merged beforehand.
                    if (counter == 254) {
                        LOG.severe(Bundle.isobus_exporter_treatment_zone_limit_exceeded());
                        break;
                    }
                    counter++;
                }
                break;
            case TYPE_2:
                // create a single prototype TZN:
                IsobusTreatmentZone isobusTreatmentZone = new IsobusTreatmentZone();
                isobusTreatmentZone.setTreatmentZoneDesignator(TREATMENT_ZONE_TYPE_2_NAME);
                isobusTreatmentZone.setTreatmentZoneCode(TREATMENTZONE_TYPE_2_CODE);
                // add all used variables to the prototype without a value (i.e. set to 0, it's a required attribute)
                // Agrosense currently only supports a single variable:
                isobusTreatmentZone.getPDV().add(IsobusObjectFactoryUtil.createProcessDataVariable(0, presentation, ddi, multiplier));
                created.add(isobusTreatmentZone);
                break;
            default:
                throw new IllegalArgumentException("Unsupported isobus grid type: " + gridType);
        }
        return created;
    }

    /**
     * Generate a Gridfilename using the specified integer. For example:
     * GRD00005 Gridfilename stored in taskdata.xml should not have an extension
     * according to the ISOBUS standard.
     *
     * Protected for testing pusposes only.
     *
     * @param counter
     * @return Generated gridfilename
     */
    protected String buildGridFileName(int counter) {
        String counterValue = String.valueOf(counter++);
        StringBuilder sb = new StringBuilder(GRID_FILE_PREFIX);
        sb.append(StringUtils.leftPad(counterValue, GRID_ID_LENGTH, '0'));
        return sb.toString();
    }

    /**
     *
     *
     * public static final DDI DEFAULT_DDI = DDI.DDI_141; // fall back to on/off
     * value which was previously hardcoded for all exports.
     *
     * , DDI ddi, double multiplier, Callback<Object> callback private final DDI
     * ddi; private final double multiplier; private Callback<Object> callback;
     * public static final double DEFAULT_MULTIPLIER = 1.0;
     *
     */
    private class TaskExportCallable implements Callable {

        private final File selectedFile;
        private final ExportableTask[] tasks;
        private final DDI ddi;
        private final double multiplier;
        private final double defaultTznValue;

        public TaskExportCallable(File selectedFile, DDI ddi, double multiplier, Double defaultTznValue, ExportableTask... tasks) {
            this.selectedFile = selectedFile;
            this.tasks = tasks;
            this.ddi = ddi;
            this.multiplier = multiplier;
            this.defaultTznValue = defaultTznValue == null ? 0.0 : defaultTznValue;
        }

        @Override
        public Object call() throws Exception {
            FileObject selectedLocation = FileUtil.toFileObject(selectedFile);
            FileObject targetLocation = getTargetLocation(selectedLocation);
            if (clearTargetContents(targetLocation)) {
                GridTypeEnum exportType = GridTypeEnum.valueOf(IsobusPreference.EXPORT_TYPE.getValue());

                ProgressHandle progress = ProgressHandleFactory.createHandle(Bundle.isobus_exporter_generating_isobus_export());
                progress.start();
                IsobusTaskData isobusTaskData = createIsobusTaskData();
                // current hardcoded VPN defaults apply to on/off values
                // do not add them for other DDIs since that will cause wrong display values in GeoControl.
                ValuePresentation presentation = ddi == DDI.DDI_141 ? createValuePresentation() : null;
                if (presentation != null) {
                    isobusTaskData.getVPN().add(presentation);
                }

                int gridCounter = 0;
                // loop through all tasks and create isobus tasks
                for (ExportableTask task : tasks) {
                    ActivityField activityField = task.getLookup().lookup(ActivityField.class);
                    IsobusTask isobusTask = createIsobusTask(task.getURI().toString());
                    // Task ID:
                    // format (TSK|TSK-)([0-9])+
                    // length 4..14
                    // 0 is not allowed, negative IDs are for records generated 
                    // on the MICS, so start from TSK1:
                    isobusTask.setTaskId(ISOBUS_TASK_ID_PREFIX + (gridCounter + 1));

                    DataLog dataLog = createDataLog();
                    isobusTask.getDLT().add(dataLog);

                    IsobusTreatmentZone emptyTreatmentZone = createEmptyTreatmentZone();
                    emptyTreatmentZone.getPDV().add(createProcessDataVariable(presentation, ddi));
                    isobusTask.getTZN().add(emptyTreatmentZone);
                    
                    IsobusTreatmentZone defaultTreatmentZone = createDefaultTreatmentZone();
                    // TODO: set PDV value to activityField average (possibly weighted using TZN areas);
                    // that wouldn't make sense for the on/off DDI though
                    defaultTreatmentZone.getPDV().add(IsobusObjectFactoryUtil.createProcessDataVariable(defaultTznValue, presentation, ddi, multiplier));
                    isobusTask.getTZN().add(defaultTreatmentZone);
                    isobusTask.setDefaultTreatmentZoneCode(defaultTreatmentZone.getTreatmentZoneCode());
                    
                    isobusTask.setTaskDesignator(task.getName());

                    // TSK
                    isobusTaskData.getTSK().add(isobusTask);

                    TreatmentZoneGrid grid = generateTreatmentZoneGrid(activityField, task.getWorkWidth());

                    isobusTask.getTZN().addAll(createTreatmentZones(activityField.getTreatmentZones(), presentation, grid, exportType, ddi, multiplier));

                    // Gridfilename should be placed in taskdata without extension
                    String gridFileName = buildGridFileName(gridCounter);

                    IsobusGrid isobusGrid = GridGenerator.createIsobusGrid(isobusTask.getTZN(), grid, exportType, gridFileName);
                    isobusTask.setGRD(isobusGrid);

                    //Start write the grid file
                    GridWriter gridWriter = new GridWriter();
                    // remove old file first if it exists:
                    FileObject gridFile = targetLocation.getFileObject(isobusGrid.getFilename(), GRID_FILE_EXT);
                    if (gridFile != null) {
                        gridFile.delete();
                    }
                    gridFile = targetLocation.createData(isobusGrid.getFilename(), GRID_FILE_EXT);
                    gridWriter.write(gridFile, exportType, grid, ddi, multiplier);
                    //End write the grid file

                    //raise the grid counter so the next task
                    //will have its own grid file
                    gridCounter++;
                }
                //FIXME needs some heavy refactoring! why pass on grid file separately? its in the task
                IsobusMarshaller isobusMarshaller = new IsobusMarshaller();
                isobusMarshaller.marshall(targetLocation, isobusTaskData);
                progress.finish();

                NotificationDisplayer.getDefault().notify(Bundle.isobus_exporter_generating_export_done_title(), INFO_ICON, Bundle.isobus_exporter_generating_isobus_full_description(), null);

            }

            return null;
        }
    }

    @Override
    public Future exportTask(final Map<String, Object> exportProperties, final ExportableTask... tasks) {
        FileChooserBuilder builder = new FileChooserBuilder(ExportAction.class);
        builder.setTitle(Bundle.isobus_exporter_choose_location());

        final JFileChooser chooser = builder.createFileChooser();
        chooser.setMultiSelectionEnabled(false);
        chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
        chooser.setFileHidingEnabled(true);

        int returnVal = chooser.showSaveDialog(null);
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            File selectedFile = chooser.getSelectedFile();
            if (selectedFile != null) {
                if (!selectedFile.exists() && !selectedFile.mkdirs()) {
                    Exceptions.printStackTrace(new RuntimeException("Could not create export folder " + selectedFile.getPath()));
                }
                ExecutorService executor = Executors.newSingleThreadExecutor();

                DDI ddi = exportProperties.containsKey(PROP_DDI) ? (DDI) exportProperties.get(PROP_DDI) : DDI.DDI_141;
                double multiplier = exportProperties.containsKey(PROP_MULTIPLIER) ? (double) exportProperties.get(PROP_MULTIPLIER) : 1.0;
                Double defaultTznValue = exportProperties.containsKey(PROP_DEFAULT_TZN_VALUE) ? (Double) exportProperties.get(PROP_DEFAULT_TZN_VALUE) : 0.0;
                return executor.submit(new TaskExportCallable(selectedFile, ddi, multiplier,defaultTznValue, tasks));
            }
        }
        return null;
    }
    private static final String EXPORT_FOLDER_NAME = "TASKDATA";

    private FileObject getTargetLocation(FileObject selectedLocation) {
        if (selectedLocation.getName().equals(EXPORT_FOLDER_NAME)) {
            return selectedLocation;
        }

        FileObject subFolder = selectedLocation.getFileObject(EXPORT_FOLDER_NAME);
        if (subFolder == null) {
            try {
                subFolder = selectedLocation.createFolder(EXPORT_FOLDER_NAME);
            } catch (IOException ex) {
                throw new RuntimeException("Could not create export subfolder", ex);
            }
        }
        return subFolder;
    }

    private boolean clearTargetContents(FileObject targetLocation) {
        if (targetLocation.getChildren().length > 0) {
            // TODO: confirm overwrite?
            for (FileObject fob : targetLocation.getChildren()) {
                try {
                    fob.delete();
                } catch (IOException ex) {
                    Exceptions.printStackTrace(ex);
                    return false;
                }
            }
        }
        return true;
    }
    /**
     * This method extracts a subList of treatment zones from the passed list.
     * The subList has a maximum size equal to the defined maximum number of
     * treatment zones supported by the current ISOBUS implementation.
     *
     * @param treatmentZones The raw list of treatment zones the user wishes to
     * export
     * @return a subList of treatment zones no greater than the treatment zone
     * limit defined in this class
     */
    public static final int TREATMENT_ZONE_LIMIT = 254;

    protected List<TreatmentZone> getMaximumNumberOfTreatmentZones(List<TreatmentZone> treatmentZones) {
        GridTypeEnum exportType = GridTypeEnum.valueOf(IsobusPreference.EXPORT_TYPE.getValue());
        switch (exportType) {
            case TYPE_2:
                return treatmentZones;
            default:
                int subListEnd = TREATMENT_ZONE_LIMIT;
                if (treatmentZones.size() < TREATMENT_ZONE_LIMIT) {
                    return treatmentZones;
                }
                return treatmentZones.subList(0, subListEnd);
        }

    }

    private TreatmentZoneGrid generateTreatmentZoneGrid(ActivityField activityField, double workWidth) {
        List<TreatmentZone> treatmentZones = activityField.getTreatmentZones();
        //FIXME: ISOBUS supports a maximum number of 255  treatment zones 
        // (including the initial empty one). In order to support more 
        // treatment zones they should probably be merged beforehand.
        List<TreatmentZone> maximumNumberOfTreatmentZones = getMaximumNumberOfTreatmentZones(treatmentZones);

        final double LATITUDE_WORKWIDTH_BEARING = 0;
        final double LONGITUDE_WORKWIDTH_BEARING = 90;
        // Create envelope of treatmentzones for this task.
        Geometry envelope = GeometryUtil.getEnvelope(maximumNumberOfTreatmentZones);

        // Calculate nr. of gridCells in width and height
        double envelopeMinX = GeometryUtil.getMinX(envelope);
        double envelopeMaxX = GeometryUtil.getMaxX(envelope);
        double envelopeWidth = envelopeMaxX - envelopeMinX;
        double envelopeMinY = GeometryUtil.getMinY(envelope);
        double envelopeMaxY = GeometryUtil.getMaxY(envelope);
        double envelopeHeight = envelopeMaxY - envelopeMinY;

        // Calculate nr of columns and rows
        Point2D latitudeWorkWidth = GeometryTools.getDestinationPoint(envelopeMinY, envelopeMinX, LATITUDE_WORKWIDTH_BEARING, workWidth);
        Point2D longitudeWorkWidth = GeometryTools.getDestinationPoint(envelopeMinY, envelopeMinX, LONGITUDE_WORKWIDTH_BEARING, workWidth);
        //switched x and y coordinate is the other way round
        double workWidthDegrees = longitudeWorkWidth.getX() - envelopeMinX;
        double workheightDegrees = latitudeWorkWidth.getY() - envelopeMinY;

        double columns = envelopeWidth / workWidthDegrees;
        double rows = envelopeHeight / workheightDegrees;

        int nrOfGridColumns = new Double(Math.ceil(columns)).intValue();
        int nrOfGridRows = new Double(Math.ceil(rows)).intValue();


        TreatmentZoneGrid grid = GridGenerator.generate(
                GeometryUtil.getCellWidth(envelope, nrOfGridColumns),
                GeometryUtil.getCellHeight(envelope, nrOfGridRows),
                GeometryUtil.getMinX(envelope),
                GeometryUtil.getMinY(envelope),
                nrOfGridRows,
                nrOfGridColumns,
                maximumNumberOfTreatmentZones);

        return grid;
    }

    @Override
    public GridCoverage2D generateExportPreviewGrid(ExportableTask task) {
        return generateTreatmentZoneGrid(task.getActivityField(), task.getWorkWidth()).getGridCoverage2D();
    }
}
